/*

  This file is part of GESTURES.DLL source code, see MAIN.C for
  detailed information.

  Copyright (C) 2000-2001 Pavel Vitis
  pavel.vitis@seznam.cz
  http://lsg.come.to/


  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

/****************************************************************************
****************************************************************************/

#include <windows.h>
#include <memory.h>
#include "main.h"
#include "lsapi.h"

EX_WIN *ex_win_list = NULL;
UINT ex_win_cnt = 0;

// Get ULONG number from string, set parsing_error and display
// error message if some error occured.
// Somewhat multi-purpose function, not too healthy C technique...
void errorBox( LPCTSTR hs, LPCTSTR prm )
{
	char msgbuf[256];

	sprintf(msgbuf, "stroke: \"%s\"\x0Dparameter: \"%s\"\x0DUnknown parameter", hs, prm);
	MessageBox( hDesktop, msgbuf, ST_NAME " parsing error", MB_OK | MB_ICONEXCLAMATION | MB_TOPMOST );
}

// Build list of excluded windows
void loadExcludedWindows(void)
{
	FILE * f;
	unsigned int i;
	PTSTR token[4];
	char buffer[MAX_LINE_LENGTH];
	if (ex_win_list)
		free(ex_win_list);
	ex_win_list = NULL;
	ex_win_cnt = 0;

	f = LCOpen(NULL);
	if (!f)
		return;
	for (i = 0; i < 4; i++)
		token[i] = malloc(MAX_LINE_LENGTH);
	while (LCReadNextConfig (f, "*StrokeExcludeWin", buffer, sizeof(buffer)))
	{
		int count;
		count = LCTokenize (buffer, token, 4, NULL);
		if (count == 4)
		{
			if (!ex_win_list)
				ex_win_list = (EX_WIN *)malloc(sizeof(EX_WIN));
			else
				ex_win_list = (EX_WIN *)realloc(ex_win_list, (ex_win_cnt+1)*sizeof(EX_WIN));

			int length = strlen(token[1]);
			ex_win_list[ex_win_cnt].module = malloc(length+1);
			strncpy(ex_win_list[ex_win_cnt].module, token[1], length+1);
			ex_win_list[ex_win_cnt].module[length] = 0;

			length = strlen(token[2]);
			ex_win_list[ex_win_cnt].class = malloc(length+1);
			strncpy(ex_win_list[ex_win_cnt].class, token[2], length+1);
			ex_win_list[ex_win_cnt].class[length] = 0;

			length = strlen(token[3]);
			ex_win_list[ex_win_cnt].title = malloc(length+1);
			strncpy(ex_win_list[ex_win_cnt].title, token[3], length+1);
			ex_win_list[ex_win_cnt].title[length] = 0;

			LOG("Window exclusion loaded:");
			LOGF("  module: '%s'", ex_win_list[ex_win_cnt].module);
			LOGF("  class : '%s'", ex_win_list[ex_win_cnt].class);
			LOGF("  title : '%s'", ex_win_list[ex_win_cnt].title);

			ex_win_cnt++;
		}
	}
	LCClose(f);
	for (i = 0; i < 3; i++)
		free(token[i]);
	return;
}

// Reverse stroke sequence in Y-axis
void reverse( LPSTR seq )
{
	if (seq == NULL)
		return;

	for (int i = 0; i < strlen(seq); i++)
	{
		switch (seq[i])
		{
			case '7':
				seq[i] = '1';
				break;
			case '8':
				seq[i] = '2';
				break;
			case '9':
				seq[i] = '3';
				break;
			case '1':
				seq[i] = '7';
				break;
			case '2':
				seq[i] = '8';
				break;
			case '3':
				seq[i] = '9';
				break;
		}
	}
}

// Shift off-centered strokes into the center of matrix
void shift_to_center( LPSTR seq )
{
	if ((seq == NULL) || (strlen(seq) < 4))
		return;

	LPSTR p;

	while ( ((p = strstr(seq, "741 ")) || (p = strstr(seq, "963 "))) && ((p == seq) || (p[-1] == ' ')) )
	{
		p[0] = '8';
		p[1] = '5';
		p[2] = '2';
	}

	while ( ((p = strstr(seq, "147 ")) || (p = strstr(seq, "369 "))) && ( (p == seq) || (p[-1] == ' ')) )
	{
		p[0] = '2';
		p[1] = '5';
		p[2] = '8';
	}

	while ( ((p = strstr(seq, "123 ")) || (p = strstr(seq, "789 "))) && ( (p == seq) || (p[-1] == ' ')) )
	{
		p[0] = '4';
		p[1] = '5';
		p[2] = '6';
	}

	while ( ((p = strstr(seq, "321 ")) || (p = strstr(seq, "987 "))) && ( (p == seq) || (p[-1] == ' ')) )
	{
		p[0] = '6';
		p[1] = '5';
		p[2] = '4';
	}
}

// Load strokes from step.rc
BOOL loadStrokes( LPCTSTR szPath )
{
	char buffer[MAX_LINE_LENGTH];
	FILE* f;
	BOOL parsing_error = FALSE;

	f = LCOpen(NULL);
	if (!f)
	{
		char steprc[MAX_PATH+1];
		strcpy(steprc, szPath);
		if ( steprc[strlen(steprc)-1] != '\\' )
			strcat(steprc, "\\");
		strcat(steprc, "step.rc");
		f = LCOpen(steprc);
	}
	if (f)
	{
		char token1[MAX_LINE_LENGTH], token2[MAX_LINE_LENGTH], token3[MAX_LINE_LENGTH], token4[MAX_LINE_LENGTH], extra[MAX_LINE_LENGTH];
		char* tokens[4];
		int count = 0;

		tokens[0] = token1;
		tokens[1] = token2;
		tokens[2] = token3;
		tokens[3] = token4;
		buffer[0] = 0;

		while ( !parsing_error && (LCReadNextConfig(f, TOKEN, buffer, sizeof(buffer))) )
		{
			PSTROKE tstroke;

			tstroke = newStroke();

			token1[0] = token2[0] = token3[0] = token4[0] = extra[0] = '\0';
			(void) LCTokenize (buffer, tokens, 2, extra);

			// Ignore lines beginning with TOKEN, that are found too. For instance "*strokeExcludeWin".
			// We need just "*stroke" now.
			if (strcmpi(TOKEN, token1))
			{
				LOGF("Skipped: '%s'", token1);
				continue;
			}

			tstroke->name = strdup(token2);

			strtrim(extra);
			tstroke->sequence = malloc(strlen(extra)+2+1);
			tstroke->sequence[0] = ' ';
			tstroke->sequence[1] = 0;
		    strcat(tstroke->sequence, extra);
			if (reverse_sequence)
				reverse(tstroke->sequence);
		    strcat(tstroke->sequence, " ");
			shift_to_center(tstroke->sequence);

			LOGF("stroke def start: %s", token2);
			LOGF("  sequence: '%s'", tstroke->sequence);


			char prm1[MAX_LINE_LENGTH], prm2[MAX_LINE_LENGTH], prm3[MAX_LINE_LENGTH], prm4[MAX_LINE_LENGTH], prmextra[MAX_LINE_LENGTH];
			char* prms[4];
//			int prmcount;

			prms[0] = prm1;
			prms[1] = prm2;
			prms[2] = prm3;
			prms[3] = prm4;
			buffer[0] = 0;

			while ( LCReadNextConfig(f, TOKEN, buffer, sizeof(buffer)) )
			{
				prm1[0] = prm2[0] = prm3[0] = prm4[0] = prmextra[0] = '\0';

				(void) LCTokenize (buffer, prms, 3, prmextra);

//				strtrim(prm1);

				if ( (strlen(prm2) == 0) || (prm2[0] == '~') )
				{
					LOGF("stroke def end: %s", tstroke->name);
					addStroke(tstroke);
					tstroke = NULL;
					count++;
					break;
				}
				else
				if ( stricmp("action", prm2) == 0 )
				{
					LOGF("  parameter: action: %s", prm3);
					setAction( &tstroke->action, prm3, prmextra);
				}
				else
				if ( stricmp("timeout", prm2) == 0 )
				{
					LOGF("  parameter: timeout: %s", prm3);
					tstroke->timeout = atoi(prm3);
				}
				else
				if ( stricmp("exclude_win", prm2) == 0 )
				{
					LOGF("  parameter: exclude_win: %s", prm3);
					tstroke->exclude_win = strdup(prm3);
				}
				else
				if ( stricmp("exclude_mod", prm2) == 0 )
				{
					LOGF("  parameter: exclude_mod: %s", prm3);
					tstroke->exclude_mod = strdup(prm3);
				}
				else
				if ( stricmp("notify_snd", prm2) == 0 )
				{
					LOGF("  parameter: notify_snd: %s", prm3);
					tstroke->notify_snd = strdup(prm3);
				}
				else
				if ( stricmp("timeout_snd", prm2) == 0 )
				{
					LOGF("  parameter: timeout_snd: %s", prm3);
					tstroke->timeout_snd = strdup(prm3);
				}

				else
				{
					errorBox(tstroke->name, prm2);
					parsing_error = TRUE;
					break;
				}

			}

			if ( tstroke != NULL )
				deleteStroke(tstroke);
		}

		LOGF("%d strokes loaded", count);

		LCClose(f);
	}
	else
		return FALSE;

	// Return TRUE, if everything went OK
	return !parsing_error;
}
