/*

  This file is part of GESTURES.DLL source code, see MAIN.C for
  detailed information.

  Copyright (C) 2000-2001 Pavel Vitis
  pavel.vitis@seznam.cz
  http://lsg.come.to/


  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

/****************************************************************************
****************************************************************************/

#include <windows.h>
#include <mmsystem.h>
#include "main.h"
#include "./libstroke/stroke.h"
#include "hookdll.h"
#include "lsapi.h"
#include <shellapi.h>
#include "../procinfo/procinfo.h"
#include "../kmhook/hookdll.h"

// Buffer for storing stroke sequence
char sequence[MAX_SEQUENCE+2];

// Is stroke-recognition in process?
BOOL send_motion;

// Mouse button that activates stroke
UINT mbutton = MK_RBUTTON;

// Hotkey that activates stroke
UINT hotkey = 0;

// Actual shift keys state. Used when keyboard-hook is set
UINT shift_keys = 0;

// Has hotkey been pressed?
BOOL hotkey_pressed = FALSE;

HWND ToolWindow;

// active_win = window in foreground, mouse_win = window under mouse cursor
// at the time of stroke start
HWND active_win = NULL, mouse_win = NULL;

#define SK_KEYDOWN  0x8000
#define SK_SYSKEY   0x4000

// Definition of some keys, that are emulated if needed
UINT _keyAltLeft[] = { 0, VK_MENU | SK_KEYDOWN | SK_SYSKEY, VK_LEFT | SK_KEYDOWN | SK_SYSKEY, VK_LEFT | SK_SYSKEY, VK_MENU | SK_SYSKEY };
UINT _keyAltRight[] = { 0, VK_MENU | SK_KEYDOWN | SK_SYSKEY, VK_RIGHT | SK_KEYDOWN | SK_SYSKEY, VK_RIGHT | SK_SYSKEY, VK_MENU | SK_SYSKEY };
UINT _keyCtrlR[] = { 0, VK_CONTROL | SK_KEYDOWN, VK_R | SK_KEYDOWN, VK_R, VK_CONTROL };

/*
UINT _keyCtrlAltF1[] = { 0,
						 SK_KEYDOWN | VK_MENU | SK_SYSKEY,
						 SK_KEYDOWN | VK_CONTROL,
						 SK_KEYDOWN | VK_S,
						 VK_S,
						 VK_CONTROL,
						 VK_MENU | SK_SYSKEY,
						 0
					   };

UINT _keyAltTab[] = { 0,
					 SK_KEYDOWN | VK_MENU | SK_SYSKEY,
					 SK_KEYDOWN | SK_SYSKEY | VK_TAB
//					 VK_TAB | SK_SYSKEY,
//					 VK_MENU
				   };
*/

UINT nMessages[] = {
	LM_GETREVID,
	0
};

// Execute command / bang
void runParameter( LPSTR action )
{
	INT flg;
	char aCommand[MAX_LINE_LENGTH], aParams[MAX_LINE_LENGTH];
	char *args[1];

	LOG("run-param enter");
	if ( !STRVALID(action) )
		return;

	args[0] = aCommand;
    aCommand[0] = 0;
    aParams[0] = 0;
	LCTokenize(action, args, 1, aParams);

	if ( aCommand[0] == '!' )
		ParseBangCommand(GetLitestepWnd(), aCommand, aParams);
	else
	{
		// Run it hidden
		if ( aCommand[0] == '@' )
		{
			flg = 0;
			memmove(aCommand, aCommand + 1, strlen(aCommand));
		}
		else
		{
			flg = SW_SHOWDEFAULT;
/*			if ( hs->flags.maximized )
				flg = SW_SHOWMAXIMIZED;
			if ( hs->flags.minimized )
				flg = SW_SHOWMINIMIZED;
			if ( hs->flags.restore )
				flg = SW_SHOWNA;//RESTORE;
*/
		}
/*
		HWND hwnd;
		RECT mousepos, wndrect;

		if ( hs->flags.one_instance && (hwnd = getHandleFromPath(aCommand)) != NULL )
		{
			LOG("wnd is running");
			if ( hwnd != GetForegroundWindow() )
			{
				LOG("wnd is not in foreground");
				if ( IsIconic(hwnd) ) {
					LOG("wnd is iconic, resore it");
					PostMessage(hwnd, WM_SYSCOMMAND, SC_RESTORE, 0);
				}
				SetForegroundWindow(hwnd);
				if ( window_animation && GetCursorPos((LPPOINT)&mousepos.left) && GetWindowRect(hwnd, &wndrect) )
				{
					mousepos.bottom = mousepos.top;
					mousepos.right = mousepos.left;
					DrawAnimatedRects(GetDesktopWindow(), 0, &mousepos, &wndrect);
				}
			}
		}
		else
*/			ShellExecute(GetLitestepWnd(), "open", aCommand, aParams, "", flg);
	}
}


//HCURSOR oldCur = NULL;

// Try to get main window
HWND GetMainWindow( HWND wnd )
{
	while ( ( GetParent(wnd) != NULL ) && !isAppWindow(wnd) )
	{
		LOGF("non-parent: %X", wnd);
		wnd = GetParent(wnd);
	}
	LOGF("parent: %X", wnd);

	return wnd;
}

// Send window to background
void SendBack( HWND wnd )
{
	HDWP dwp;

//	if ( hDesktop == NULL )
//		SetWindowPos( wnd, HWND_BOTTOM, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE);
//	else
	{
		dwp = BeginDeferWindowPos(1);
		dwp = DeferWindowPos(dwp, wnd, HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
		dwp = DeferWindowPos(dwp, hDesktop, HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
		EndDeferWindowPos(dwp);
	}
}

// Play sound
void sound( LPCTSTR snd, LPCTSTR def )
{
	if ( snd )
		sndPlaySound( snd, SND_ASYNC | SND_NOSTOP );
	else
	if ( def )
		sndPlaySound( def, SND_ASYNC | SND_NOSTOP );
}

// Try to cancel special cursor modes (dragging, selecting, context-menu open, etc.)
// after stroke start.
// This is very slicky and dirty, though sometimes not successful
BOOL CALLBACK cancelModeProc(HWND hwnd,	LPARAM lParam)
{
//	char classbuf[256];

//	GetClassName(hwnd, classbuf, sizeof(classbuf));
//	LOGF("Enumwnd: %s", classbuf);
	SendMessage(hwnd, WM_CANCELMODE, 0, 0);
	SendMessage(hwnd, WM_EXITSIZEMOVE, 0, 0);
	SendMessage(hwnd, EM_SETSEL, -1, 0);
//	SendMessage(hwnd, WM_CAPTURECHANGED, 0, 0);
	return TRUE;
}

void cancelModes( HWND wnd )
{
//	char classbuf[256];
	HWND twnd, twnd1;
	DWORD thid;

	twnd = wnd;
	do {
		SendMessage(twnd, WM_CANCELMODE, 0, 0);
		SendMessage(twnd, WM_CAPTURECHANGED, 0, 0);
		twnd = GetWindow(twnd, GW_HWNDPREV);
	} while ( twnd && (twnd != wnd) );

	EnumChildWindows(wnd, cancelModeProc, 0);

	if ( GetWindowThreadProcessId(wnd, &thid) != 0 )
		EnumThreadWindows(thid, cancelModeProc, 0);

	twnd = GetWindow(wnd, GW_CHILD);
	twnd1 = twnd;
	do {
		SendMessage(twnd, WM_CANCELMODE, 0, 0);
		SendMessage(twnd, WM_CAPTURECHANGED, 0, 0);
		twnd = GetWindow(twnd, GW_HWNDNEXT);
	} while ( twnd && (twnd != twnd1) );

	twnd = GetWindow(wnd, GW_HWNDFIRST);
	PostMessage(GetWindow(twnd, GW_OWNER), WM_CANCELMODE, 0, 0);
	PostMessage(GetWindow(twnd, GW_OWNER), WM_CAPTURECHANGED, 0, 0);

	PostMessage(GetWindow(wnd, GW_OWNER), WM_CANCELMODE, 0, 0);
	PostMessage(GetWindow(wnd, GW_OWNER), WM_CAPTURECHANGED, 0, 0);

	if ( twnd = GetParent(wnd) )
	{
		SendMessage(twnd, WM_CANCELMODE, 0, 0);
		SendMessage(twnd, WM_CAPTURECHANGED, 0, 0);
	}

	if ( twnd = GetCapture() )
	{
		SendMessage(twnd, WM_CANCELMODE, 0, 0);
		SendMessage(twnd, WM_CAPTURECHANGED, 0, 0);
	}

/*
	if ( GetClassName(twnd, classbuf, sizeof(classbuf)) )
	{
		if ( strcmp("#32768", classbuf) == 0 )
		{
			PostMessage(GetWindow(twnd, GW_OWNER), WM_CANCELMODE, 0, 0);
//			PostMessage(twnd, WM_CLOSE, 0, 0);
		}
	}
	*/
}

void WindowCommand(void)
{
}

typedef struct {
	LPSTR class;
	LPSTR caption;
	UINT msg;
	WPARAM wparam;
	LPARAM lparam;
} CSTRUCT, *PCSTRUCT;

static CSTRUCT cstruct;

BOOL CALLBACK enumControlsProc(HWND hwnd, LPARAM pcstruct)
{
	char buf[256];

	if ( GetClassName(hwnd, buf, sizeof(buf)) == 0 )
		return TRUE;

	LOGF("search for: \"%s\"", ((PCSTRUCT) pcstruct)->class);
	LOGF("Enumwnd: %s", buf);

/*	if ( strcmpi("MS_WINTOPIC", buf) == 0 )
	{
		SendMessage(hwnd, WM_RBUTTONUP, 0, 0);
		SendMessage(hwnd, WM_LBUTTONDOWN, 0, 0);
		SendMessage(hwnd, WM_LBUTTONUP, 0, 0);
		return TRUE;
	}
*/
	if ( strcmpi(((PCSTRUCT) pcstruct)->class, buf) != 0 )
		return TRUE;
	GetWindowText(hwnd, buf, sizeof(buf));
	LOGF("Enumwnd caption: \"%s\"", buf);
	if ( strcmpi(((PCSTRUCT) pcstruct)->caption, buf) != 0 )
		return TRUE;

	SendMessage(hwnd, ((PCSTRUCT) pcstruct)->msg, ((PCSTRUCT) pcstruct)->wparam, ((PCSTRUCT) pcstruct)->lparam);
	return FALSE;
}

// Send window message to specific window control
int SendMessageToControl(HWND wnd, LPSTR class, LPSTR caption, UINT msg, WPARAM wparam, LPARAM lparam)
{
	cstruct.class = class;
	cstruct.caption = caption;
	cstruct.msg = msg;
	cstruct.wparam = wparam;
	cstruct.lparam = lparam;
	return EnumChildWindows(wnd, &enumControlsProc, (LPARAM) &cstruct);
}

// Emulate "Back" function in Windows help window by pressing "Back" button
// This function should be replaced by SendMessageToControl in future.
void hlpback(HWND wnd)
{
	char classbuf[256];
	if ( GetClassName(wnd, classbuf, sizeof(classbuf)) != 0 )
		LOGF("classname: %s",classbuf);
	if ( !match("MS_WIN*" , classbuf) )
		return;
	SendMessageToControl(wnd, "Button", "&Back", BM_CLICK, 0, 0);
	SendMessageToControl(wnd, "Button", "&Zpt", BM_CLICK, 0, 0); // "Back" = "Zpet" in Czech version of Windows ;-)
}

volatile static POINT mousepos_on_click, mousepos_on_release;
static WPARAM wparam_on_click;
static BOOL ignore_mouse = FALSE;

#define WM_SIMULATEMOUSE WM_USER

static char class[256];
static char fname[MAX_PATH+1];

// Start recording mouse stroke
BOOL start_stroke(PMOUSEHOOKSTRUCT mhstruct, BOOL was_mbutton)
{
	// Was stroke started by mouse-click?
	if (was_mbutton)
	{
		// Yes, test for hotkeys
		if (hotkey & HK_SHIFT)
		{
			LOG("Testing for SHIFT hotkey");
			if ( !SHIFTKEY )
				return FALSE;
		}
		if (hotkey & HK_ALT)
		{
			LOG("Testing for ALT hotkey");
			if ( !ALTKEY )
				return FALSE;
		}
		if (hotkey & HK_CTRL)
		{
			LOG("Testing for CTRL hotkey");
			if ( !CTRLKEY )
				return FALSE;
		}
	}

//	save_mouse_msg = wParam;
	mousepos_on_click = mhstruct->pt;
	// Construct wParam
	wparam_on_click = MK_RBUTTON;
	if ( SHIFTKEY )
		wparam_on_click |= MK_SHIFT;
	if ( CTRLKEY )
		wparam_on_click |= MK_CONTROL;
	LOGF("MOUSE-X: %d", mousepos_on_click.x);
	LOGF("MOUSE-Y: %d", mousepos_on_click.y);
	mouse_win = mhstruct->hwnd;//WindowFromPoint( mousepos_on_click );
	LOGF("MOUSEHOOKSTRUCT extrainfo: %d", mhstruct->dwExtraInfo);
	LOGF("wnd from point: %X", mouse_win);
//				if ( MHSTRUCT->hwnd == mouse_win )
//					LOG("mouse_win and MHSTRUCT->hwnd matches!");
	class[0] = 0;
	GetClassName(mouse_win, class, sizeof(class));
	fname[0] = 0;
	getWndFilename(mouse_win, fname, sizeof(fname));
	{
//		if ( match("DesktopBackgroundClass", classbuf) )
//			break;
//		if ( match("*TWharf*", classbuf) )
//			break;
//		if ( match("*PopupMenu*", classbuf) )
//			break;
//		if ( getWndFilename(mouse_win, fname, sizeof(fname)) )
		if (!no_def_excludes)
		{
			LOGF("%s", fname);
			if ( match("*\\OPERA.EXE", fname) /*&& (stricmp("BLD_MDICHILD", classbuf) == 0)*/ )
			{
				LOG("  Excluded window: Opera");
				return FALSE;
			}
		}
/*						if (stricmp("IEFrame", classbuf) == 0)
		{
			LOG("IE window found!");
			SendMessage(mouse_win, WM_KEYDOWN, VK_F5, 0);
		}
*/
	}

	char title[256];
	title[0] = '\0';
	GetWindowText(mouse_win, title, sizeof(title));
	if (excluded_window(fname, class, title))
		return FALSE;

	active_win = GetForegroundWindow();
//				oldCur = SetCursor(LoadCursor(NULL, MAKEINTRESOURCE("IDC_CROSS")));
//					if (GetClassName(active_win, classbuf, sizeof(classbuf)))
//						if (stricmp("DesktopWindowClass", classbuf) == 0)
//							break;
	LOG("Start gesture recording");
	send_motion = TRUE;
	startTimer( def_timeout );
	startDesktopDrawing(mousepos_on_click.x, mousepos_on_click.y);

	return TRUE;
}

BOOL end_stroke(PMOUSEHOOKSTRUCT mhstruct, BOOL was_mbutton)
{
	char tempseq[MAX_SEQUENCE+2];

	mousepos_on_release = mhstruct->pt;

	if (send_motion)
	{
//		SetCursor(oldCur);
		send_motion = FALSE;
		stopTimer();
		stopDesktopDrawing();
		if ( stroke_trans(sequence) )
		{
			LOGF("Translation succeeded: %s", sequence);

			tempseq[0] = ' ';
			tempseq[1] = 0;
			strcat(tempseq, sequence);
			strcat(tempseq, " ");

			PSTROKE st;

			st = strokes;
			while ( st )
			{
				if ( strstr(st->sequence, tempseq) )
				{
					LOGF("Stroke match: '%s'", st->sequence);
					cancelModes(active_win);
					cancelModes(mouse_win);
//					cancelModes(GetMainWindow(mouse_win));
//					if ( isAppWindow(active_win) )
					{
						switch (st->action.id)
						{
							case ACT_EXECUTE:
							case ACT_BANG:
								runParameter(st->action.parm);
								break;
							case ACT_WC_CLOSE:
								if ( !no_def_excludes && ((mouse_win == hDesktop) || match("*TWharf*", class) || match("*PopupMenu*", class)
									 || match("*ShortcutClass*", class)) )
								{
									LOGF("Improper window for such operation!!! class = '%s'", class);
								}
								else
									PostMessage(GetMainWindow(mouse_win), WM_SYSCOMMAND, SC_CLOSE, 0);
								break;
							case ACT_WC_MAXIMIZE:
								if ( !no_def_excludes && ((mouse_win == hDesktop) || match("*TWharf*", class) || match("*PopupMenu*", class)
									|| match("*ShortcutClass*", class)) )
								{
									LOGF("Improper window for such operation!!! class = '%s'", class);
								}
								else
								{
									PostMessage(GetMainWindow(mouse_win), WM_SYSCOMMAND, SC_MAXIMIZE, 0);
									PostMessage(active_win, WM_SYSCOMMAND, SC_MAXIMIZE, 0);
								}
								break;
							case ACT_WC_MINIMIZE:
								if ( !no_def_excludes && ((mouse_win == hDesktop) || match("*TWharf*", class) || match("*PopupMenu*", class)
									|| match("*ShortcutClass*", class)) )
								{
									LOGF("Improper window for such operation!!! class = '%s'", class);
								}
								else
									PostMessage(GetMainWindow(mouse_win), WM_SYSCOMMAND, SC_MINIMIZE, 0);
								break;
							case ACT_WC_RESTORE:
								if ( !no_def_excludes && ((mouse_win == hDesktop) || match("*TWharf*", class) || match("*PopupMenu*", class)
									|| match("*ShortcutClass*", class)) )
								{
									LOGF("Improper window for such operation!!! class = '%s'", class);
								}
								else
									PostMessage(GetMainWindow(mouse_win), WM_SYSCOMMAND, SC_RESTORE, 0);
								break;
							case ACT_WC_SENDBACK:
								if ( !no_def_excludes && ((mouse_win == hDesktop) || match("*TWharf*", class) || match("*PopupMenu*", class)
									|| match("*ShortcutClass*", class)) )
								{
									LOG("Improper window for such operation!!!");
								}
								else
								{
//									LOG("Sending back");
									SendBack( GetMainWindow(active_win) );
//									SetWindowPos( active_win, HWND_BOTTOM, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE);
//									SetWindowPos( hDesktop, HWND_BOTTOM, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE);
//									SetForegroundWindow( hDesktop );
								}
								break;
							case ACT_HELP_BACK:
								hlpback(GetMainWindow(active_win));
								break;
							case ACT_BROWSER_BACK:
//								SetForegroundWindow(active_win);
								(sendKeys)(_keyAltLeft, sizeof(_keyAltLeft) / sizeof(UINT));
								break;
							case ACT_BROWSER_FORWARD:
//								SetForegroundWindow(active_win);
								(sendKeys)(_keyAltRight, sizeof(_keyAltRight) / sizeof(UINT));
//								(sendKeys)(_keyAltTab, sizeof(_keyAltTab) / sizeof(UINT));
								break;
							case ACT_BROWSER_RELOAD:
//								SetForegroundWindow(active_win);
								(sendKeys)(_keyCtrlR, sizeof(_keyCtrlR) / sizeof(UINT));
								break;
						}	// switch
					}	// if
					sound( st->notify_snd, def_notify_snd );
					return TRUE;
				}	// if
				st = st->next;
			}	// while
		}	// if
		else
		{
			LOG("Translation failed");
			ignore_mouse = TRUE;
			LOGF("# MOUSE-X: %d", mousepos_on_click.x);
			LOGF("# MOUSE-Y: %d", mousepos_on_click.y);
//			DWORD pid;
//			GetWindowThreadProcessId(mouse_win, &pid);
//			PostThreadMessage(pid, WM_RBUTTONDOWN, wparam_on_click, MAKELPARAM(mousepos_on_click.x, mousepos_on_click.y));
//			SendMessage(mouse_win, WM_RBUTTONDOWN, wparam_on_click, MAKELPARAM(mousepos_on_click.x, mousepos_on_click.y));
//			ignore_mouse = FALSE;
			PostMessage(ToolWindow, WM_SIMULATEMOUSE, 0, 0);
//			(sendMouseClick)(mouse_win, WM_RBUTTONDOWN, mousepos_on_click);
//			(sendMouseClick)(mouse_win, WM_RBUTTONUP, mousepos_on_click);
//			ignore_mouse = FALSE;
			LOG("Sendmouseclick end");
			return TRUE;
		}
	}

	return FALSE;
}

static MOUSEHOOKSTRUCT mhs;

LRESULT WINAPI WndProc( HWND hWnd, UINT nMessage, WPARAM wParam, LPARAM lParam )
{
	LPTSTR pszBuf;

	// Only hotkey checking, no mouse button defined
	if ( nMessage == KeyboardMsg )
	{
//		LOGF("KEY wParam: %08X", wParam);
//		LOGF("KEY lParam: %08X", lParam);

		// Is key-up?
		if (lParam & 0x80000000U)
		{
//			LOG("Key up");
			if ( wParam == VK_SHIFT )
			{
				shift_keys &= ~HK_SHIFT;
//				LOG("SHIFT OFF");
			}
			else
			if ( wParam == VK_MENU )
			{
				shift_keys &= ~HK_ALT;
//				LOG("ALT OFF");
			}
			else
			if ( wParam == VK_CONTROL )
			{
				shift_keys &= ~HK_CTRL;
//				LOG("CTRL OFF");
			}

			// Hotkey has been released
			if ( hotkey != shift_keys )
			{
				LOG("HOTKEY OFF!!!");
				hotkey_pressed = FALSE;
				end_stroke(&mhs, FALSE);
			}
		}
		else
		// Is key-down?
		// Ignore key-repeats (test for previous key state - must have been 'up').
		if ((lParam & 0x40000000U) == 0)
		{
//			LOG("Key down");
			if ( wParam == VK_SHIFT )
			{
				shift_keys |= HK_SHIFT;
//				LOG("SHIFT ON");
			}
			else
			if ( wParam == VK_MENU )
			{
				shift_keys |= HK_ALT;
//				LOG("ALT ON");
			}
			else
			if ( wParam == VK_CONTROL )
			{
				shift_keys |= HK_CTRL;
//				LOG("CTRL ON");
			}

			// Hotkey combination has been pressed
			if ( hotkey & shift_keys )
			{
				hotkey_pressed = TRUE;
				LOG("HOTKEY ON!!!");
			}
		}
		return 0;
	}

	if ( nMessage == MouseMsg )
	{
//		LOG("Gestures MouseMsg called!!!");
		if ( ignore_mouse )
		{
			LOG("MouseMsg Ignored!!!");
		}
		else
		switch (wParam)
		{
			case WM_MOUSEWHEEL:
				LOG("MOUSEWHEEL");
//				return 1;
				break;
			case WM_RBUTTONDOWN:
			case WM_NCRBUTTONDOWN:
				if (mbutton == MK_RBUTTON)
					if ( start_stroke((PMOUSEHOOKSTRUCT) lParam, TRUE) )
						return 1;
				break;
			case WM_LBUTTONDOWN:
			case WM_NCLBUTTONDOWN:
				if (mbutton == MK_LBUTTON)
					if ( start_stroke((PMOUSEHOOKSTRUCT) lParam, TRUE) )
						return 1;
				break;
			case WM_MBUTTONDOWN:
			case WM_NCMBUTTONDOWN:
				if (mbutton == MK_MBUTTON)
					if ( start_stroke((PMOUSEHOOKSTRUCT) lParam, TRUE) )
						return 1;
				break;
			case WM_RBUTTONUP:
			case WM_NCRBUTTONUP:
//				if ( !ignore_mouse )
//					ignore_mouse = FALSE;
//				else
				if (mbutton == MK_RBUTTON)
					if ( end_stroke((PMOUSEHOOKSTRUCT) lParam, TRUE) )
						return 1;
				break;
			case WM_LBUTTONUP:
			case WM_NCLBUTTONUP:
//				if ( !ignore_mouse )
//					ignore_mouse = FALSE;
//				else
				if (mbutton == MK_LBUTTON)
					if ( end_stroke((PMOUSEHOOKSTRUCT) lParam, TRUE) )
						return 1;
				break;
			case WM_MBUTTONUP:
			case WM_NCMBUTTONUP:
//				if ( !ignore_mouse )
//					ignore_mouse = FALSE;
//				else
				if (mbutton == MK_MBUTTON)
					if ( end_stroke((PMOUSEHOOKSTRUCT) lParam, TRUE) )
						return 1;
				break;
			case WM_MOUSEMOVE:
			case WM_NCMOUSEMOVE:
				// Do something only if the mouse has changed position.
				// We are getting mouse-move message even after cursor redraws,
				// when mouse is not moving actually, so be effective.
				if ( (((PMOUSEHOOKSTRUCT) lParam)->pt.x != mhs.pt.x) || (((PMOUSEHOOKSTRUCT) lParam)->pt.y != mhs.pt.y) )
				{
					// Are we in stroke-recognition phase?
					if (send_motion)
					{
						// Keep timer up
						startTimer( def_timeout );
						// Record mouse moving
						stroke_record( ((PMOUSEHOOKSTRUCT) lParam)->pt.x, ((PMOUSEHOOKSTRUCT) lParam)->pt.y );
						// Draw trace, if needed
						desktopDrawPoint( ((PMOUSEHOOKSTRUCT) lParam)->pt.x, ((PMOUSEHOOKSTRUCT) lParam)->pt.y );
	//					return 1;
					}
					else
					// Hotkey has been activated, so start stroke, because mouse is moving now
					if ( hotkey_pressed )
					{
						start_stroke((PMOUSEHOOKSTRUCT) lParam, FALSE);
					}
				}
				break;
		}
		if ((PMOUSEHOOKSTRUCT) lParam != NULL)
			memcpy(&mhs, (PMOUSEHOOKSTRUCT) lParam, sizeof(MOUSEHOOKSTRUCT));
		return 0;
	}

	switch( nMessage )
	{
		case WM_SIMULATEMOUSE:
			if (!hotkey_pressed)
			{
				LOG("MOUSE SIMULATED");
//				ignore_mouse = TRUE;
				(sendMouseClick)(mouse_win,
								 mbutton == MK_LBUTTON ? WM_LBUTTONDOWN :
								 mbutton == MK_RBUTTON ? WM_RBUTTONDOWN :
								 WM_MBUTTONDOWN , mousepos_on_click);
//				sleep(1000);
				(sendMouseClick)(mouse_win,
								 mbutton == MK_LBUTTON ? WM_LBUTTONUP :
								 mbutton == MK_RBUTTON ? WM_RBUTTONUP :
								 WM_MBUTTONUP, mousepos_on_release);
			}
			else
				hotkey_pressed = FALSE;
			ignore_mouse = FALSE;
			LOG("MOUSE SIMULATED OK");
			break;
		case WM_TIMER:
			// Stroke timeout passed
			send_motion = FALSE;
			stopDesktopDrawing();
			stopTimer();
			sound( NULL, def_timeout_snd );
			LOG("Timeout passed");
			ignore_mouse = TRUE;
//			DWORD pid;
//			GetWindowThreadProcessId(mouse_win, &pid);
//			PostThreadMessage(pid, WM_RBUTTONDOWN, wparam_on_click, MAKELPARAM(mousepos_on_click.x, mousepos_on_click.y));
//			SendMessage(mouse_win, WM_RBUTTONDOWN, wparam_on_click, MAKELPARAM(mousepos_on_click.x, mousepos_on_click.y));
			if (!hotkey_pressed)
			{
				POINT mp;
				GetCursorPos(&mp);
				(sendMouseClick)(mouse_win,
								 mbutton == MK_LBUTTON ? WM_LBUTTONDOWN :
								 mbutton == MK_RBUTTON ? WM_RBUTTONDOWN :
								 WM_MBUTTONDOWN, mp);
			}
			else
				hotkey_pressed = FALSE;
			ignore_mouse = FALSE;
			return 0;
			break;
		case WM_CREATE:     // window create ok
			return 0;
			break;
		case WM_ERASEBKGND: // don't erase background
			return 0;
			break;
		case WM_PAINT:
			return 0;
			break;
		case WM_KEYDOWN:    // forward keyboard messages to parent window
//			PostMessage(parent,msg,wParam,lParam);
//	result:=0;
			break;
		case WM_KEYUP:
//			PostMessage(parent,msg,wParam,lParam);
//	result:=0;
			break;
		case LM_GETREVID:
			pszBuf = (LPTSTR) lParam;
			if( wParam == 0 || wParam == 1 )
				strcpy( pszBuf, ST_REVID );
			else
				pszBuf[0] = 0;
			return strlen( pszBuf );
			break;
	}

	return DefWindowProc( hWnd, nMessage, wParam, lParam );
}

BOOL createToolWindow(void)
{
	WNDCLASS wc;

	// register main window class
	memset(&wc, '\0', sizeof(wc));
//	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_DBLCLKS | CS_NOCLOSE;;
	wc.lpfnWndProc = (WNDPROC) WndProc;
//	wc.cbClsExtra = 0;
//	wc.cbWndExtra = 0;
	wc.hInstance = DLLInstance;
///	wc.hCursor = LoadCursor( NULL, IDC_ARROW );
///	wc.hIcon = NULL;
//	wc.hbrBackground = NULL;
//	wc.lpszMenuName = NULL;
	wc.lpszClassName = WC_MAIN;
//	wc.hIconSm = NULL;

	if (RegisterClass(&wc) == 0)
		return FALSE;

	// create main module window
	ToolWindow = CreateWindowEx(
		WS_EX_NOPARENTNOTIFY | WS_EX_TOOLWINDOW | WS_EX_TRANSPARENT,
		WC_MAIN,
		NULL,
		WS_POPUP,
		0,
		0,
		0,
		0,
		NULL,
		0,
		DLLInstance,
		NULL );

	if (ToolWindow == NULL)
	{
		UnregisterClass( WC_MAIN, DLLInstance );
		return FALSE;
	}

	// hide us from the VWM and set initial visibility state
	SetWindowLong( ToolWindow, GWL_USERDATA, magicDWord );
	SetParent( ToolWindow, hLitestep );

	// register for messages from the LS core
	SendMessage( hLitestep, LM_REGISTERMESSAGE, (WPARAM) ToolWindow, (LPARAM) nMessages );

	return TRUE;
}

void destroyToolWindow(void)
{
	SendMessage( hLitestep, LM_UNREGISTERMESSAGE, (WPARAM) ToolWindow, (LPARAM) nMessages );
	DestroyWindow( ToolWindow );
	UnregisterClass( WC_MAIN, DLLInstance );
}

