/*
  This file is part of KMHOOK.DLL source code
  KMHOOK.DLL is native Win32 library providing system-wide hooks

  Copyright (C) 2000-2001 Pavel Vitis
  pavel.vitis@seznam.cz
  http://lsg.come.to/

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include "../kmhook/hookdll.h"

//#define BE_SAFE

#ifdef HOOK_LOG

	#include <stdio.h>

	char logmsg[256];
	LPVOID lpMsgBuf;

	#define LOG(TXT)		OutputDebugString(HOOK_DLL TEXT(": ") TXT TEXT("\n"))
	#define LOGF(TXT,PRM)	sprintf(logmsg, HOOK_DLL TEXT(": ") TXT TEXT("\n"), PRM); \
							OutputDebugString(logmsg)
	#define LOGIF(IF,TXT)	if (IF)	OutputDebugString(HOOK_DLL TEXT(": ") TXT TEXT("\n"))
	#define LOGLE			FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM, NULL, GetLastError(), \
							MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &lpMsgBuf, 0, NULL); \
							OutputDebugString(lpMsgBuf); \
							LocalFree( lpMsgBuf )
#else
	#define LOG(TXT)
	#define LOGF(TXT,PRM)
	#define LOGIF(IF,TXT)
	#define LOGLE
#endif

UINT MouseMsg = 0;
UINT KeyboardMsg = 0;
UINT ShellMsg = 0;
PSHDATASTRUCT SharedMem = NULL;
HMODULE hLibHookLib = NULL;

GETMSGPROC getMonitorKeyboardMsg, getMonitorMouseMsg, getMonitorShellMsg;
SETHOOKPROC setMouseHook, setKeyboardHook, setShellHook;
UNSETHOOKPROC unsetMouseHook, unsetKeyboardHook, unsetShellHook;
SENDMOUSECLICKPROC sendMouseClick;
SENDKEYSPROC sendKeys;
BOOLPROC isHookSet;
WORDPROC getHookDLLVersion;
GETSHAREDMEM getSharedMem;

#ifdef BE_SAFE
// Define some dummy functions for case of DLL laoding error

__cdecl WORD dummyGetDLLVersion( void )
{
	return 0x0100;
}

__cdecl BOOL dummySetHook( HWND wnd, EVENTS evnt )
{
	return FALSE;
}

__cdecl BOOL dummyUnsetHook( HWND wnd )
{
	return FALSE;
}

__cdecl BOOL dummyIsHookSet( void )
{
	return FALSE;
}

__cdecl UINT dummyGetMsg( void )
{
	return 0;
}
#endif // BE_SAFE

// Dynamically load hook dll into address space

BOOL loadHookDLL( HINSTANCE hInstance, WORD need_version )
{
	char DLLPath[MAX_PATH+1];
	char DLLDrive[_MAX_DRIVE+1];
	char DLLDir[_MAX_DIR+1];
	char msgbuf[256];

	// Try to get hook dll handle, if already in memory
	hLibHookLib = GetModuleHandle(HOOK_DLL);
	if ( hLibHookLib )
	{
		GetModuleFileName( hLibHookLib, DLLPath, MAX_PATH );
		LOGF("%s", DLLPath);
	}
	else
	{
		// Load it from disk
		GetModuleFileName( hInstance, DLLPath, sizeof(DLLPath) );
		_splitpath( DLLPath, DLLDrive, DLLDir, NULL, NULL );
		strcpy( DLLPath, DLLDrive );
		strcat( DLLPath, DLLDir );
		strcat( DLLPath, HOOK_DLL );
		LOGF("%s", DLLPath);
		hLibHookLib = LoadLibrary(DLLPath);
	}

	if ( hLibHookLib )
    {
		// Get pointers to functions

		getSharedMem = (GETSHAREDMEM) GetProcAddress( hLibHookLib, "getMemAddress");
		if ( getSharedMem != NULL )
		{
			SharedMem = (getSharedMem)();
			LOGF("Shared MEM: %d", SharedMem);
		}

		sendMouseClick = (SENDMOUSECLICKPROC) GetProcAddress( hLibHookLib, "sendMouseClick");
		sendKeys = (SENDKEYSPROC) GetProcAddress( hLibHookLib, "sendKeys");

		getMonitorMouseMsg = (GETMSGPROC) GetProcAddress( hLibHookLib, "getMonitorMouseMsg");
		if ( getMonitorMouseMsg != NULL )
		{
			MouseMsg = (getMonitorMouseMsg)();
			LOGF("MouseMsg: %d", MouseMsg);
		}
#ifdef BE_SAFE
		else
			getMonitorMouseMsg = dummyGetMsg;
#endif // BE_SAFE

        getMonitorKeyboardMsg = (GETMSGPROC) GetProcAddress(hLibHookLib, "getMonitorKeyboardMsg");
        if ( getMonitorKeyboardMsg != NULL )
		{
			KeyboardMsg = (getMonitorKeyboardMsg)();
			LOGF("KeyboardMsg: %d", KeyboardMsg);
		}
#ifdef BE_SAFE
		else
			getMonitorKeyboardMsg = dummyGetMsg;
#endif // BE_SAFE

        getMonitorShellMsg = (GETMSGPROC) GetProcAddress(hLibHookLib, "getMonitorShellMsg");
        if ( getMonitorShellMsg != NULL )
		{
			ShellMsg = (getMonitorShellMsg)();
			LOGF("ShellMsg: %d", ShellMsg);
		}
#ifdef BE_SAFE
		else
			getMonitorShellMsg = dummyGetMsg;
#endif // BE_SAFE

		setKeyboardHook = (SETHOOKPROC) GetProcAddress(hLibHookLib, "setKeyboardHook");
#ifdef BE_SAFE
		if ( setKeyboardHook == NULL )
			setKeyboardHook = dummySetHook;
#endif // BE_SAFE

        unsetKeyboardHook = (UNSETHOOKPROC) GetProcAddress(hLibHookLib, "unsetKeyboardHook");
#ifdef BE_SAFE
		if ( unsetKeyboardHook == NULL )
			unsetKeyboardHook = dummyUnsetHook;
#endif // BE_SAFE

		setMouseHook = (SETHOOKPROC) GetProcAddress(hLibHookLib, "setMouseHook");
#ifdef BE_SAFE
		if ( setMouseHook == NULL )
			setMouseHook = dummySetHook;
#endif // BE_SAFE

        unsetMouseHook = (UNSETHOOKPROC) GetProcAddress(hLibHookLib, "unsetMouseHook");
#ifdef BE_SAFE
		if ( unsetMouseHook == NULL )
			unsetMouseHook = dummyUnsetHook;
#endif // BE_SAFE

		setShellHook = (SETHOOKPROC) GetProcAddress(hLibHookLib, "setShellHook");
#ifdef BE_SAFE
		if ( setShellHook == NULL )
			setShellHook = dummySetHook;
#endif // BE_SAFE

        unsetShellHook = (UNSETHOOKPROC) GetProcAddress(hLibHookLib, "unsetShellHook");
#ifdef BE_SAFE
		if ( unsetShellHook == NULL )
			unsetShellHook = dummyUnsetHook;
#endif // BE_SAFE

		isHookSet = (BOOLPROC) GetProcAddress(hLibHookLib, "isHookSet");
#ifdef BE_SAFE
		if ( isHookSet == NULL )
			isHookSet = dummyIsHookSet;
#endif // BE_SAFE

		getHookDLLVersion = (WORDPROC) GetProcAddress(hLibHookLib, "getHookDLLVersion");
#ifdef BE_SAFE
		if ( getHookDLLVersion == NULL )
			getHookDLLVersion = dummyGetDLLVersion;
#endif // BE_SAFE

	}
	else
		MessageBox(NULL, "Could not load", HOOK_DLL TEXT(" error"), MB_OK | MB_ICONEXCLAMATION | MB_TOPMOST);

	WORD curver = getHookDLLVersion();
	LOGF("version: %04x", curver);
	if ( need_version > curver )
	{
		LOGF("Old version, need at least %04x", need_version);
		sprintf(msgbuf, HOOK_DLL " version %2d.%02d found.\nNeeds version %2d.%02d or above.",
			HIBYTE(curver), LOBYTE(curver), HIBYTE(need_version), LOBYTE(need_version));
		MessageBox(NULL, msgbuf, HOOK_DLL TEXT(" error"), MB_OK | MB_ICONEXCLAMATION | MB_TOPMOST);
		FreeLibrary(hLibHookLib);
		hLibHookLib = NULL;
	}

	return (hLibHookLib != NULL);
}

void unloadHookDLL( void )
{
	if ( !isHookSet() )
	{
		FreeLibrary(hLibHookLib);
		#ifdef HOOK_LOG
    	OutputDebugString( HOOK_DLL TEXT(" Library unloaded") );
		#endif
	}
}

